/*
 * Copyright (c) 2009 The openGion Project.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.opengion.plugin.io;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Locale;

import org.apache.poi.ss.usermodel.Cell;
import org.apache.poi.ss.usermodel.CreationHelper;
import org.apache.poi.ss.usermodel.Font;
import org.apache.poi.ss.usermodel.RichTextString;
import org.apache.poi.ss.usermodel.Row;
import org.apache.poi.ss.usermodel.Sheet;
import org.apache.poi.ss.usermodel.Workbook;
import org.apache.poi.ss.usermodel.WorkbookFactory;

import org.opengion.fukurou.model.NativeType;
import org.opengion.fukurou.util.Closer;
import org.opengion.fukurou.util.StringUtil;
import org.opengion.hayabusa.common.HybsSystem;
import org.opengion.hayabusa.common.HybsSystemException;
import org.opengion.hayabusa.db.DBTableModel;

/**
 * ネイティブEXCELファイルの書き出しクラスです。
 *
 * DefaultTableWriter を継承していますので，ラベル，名前，データの出力部のみ
 * オーバーライドして，MIcrosoft Excelファイルの出力機能を実現しています。
 *
 * 出力形式は、openXML形式にも対応しています。
 * 出力ファイルの拡張子が、.xlsならExcel2003のバイナリ形式、.xlsxならExcel2007の
 * openXML形式で出力されます。
 *
 * @og.group ファイル出力
 *
 * @og.rev 4.3.4.3 (2008/12/22) 一部protected化
 * @og.rev 4.3.6.7 (2009/05/22) ooxml形式対応
 * @og.rev 6.0.1.2 (2014/08/08) ロジックの完全書き換え(保証もできない位･･･)
 *
 * @version  4.0
 * @author	 Kazuhiko Hasegawa
 * @since    JDK5.0,
 */
public class TableWriter_Excel extends TableWriter_Default {
	//* このプログラムのVERSION文字列を設定します。	{@value} */
	private static final String VERSION = "6.0.1.2 (2014/08/08)" ;

	private Sheet	sheet			= null;
	private	int		nRowIndex		= 0;				// 6.0.1.2 (2014/08/08) private化
	private String  filename		= null;				// 3.5.4.3 (2004/01/05)
	private String  sheetName		= "Sheet1";			// 3.5.4.3 (2004/01/05)
	private String	fontName		= null;				// 3.8.5.3 (2006/08/07)
	private short	fontPoint		= -1;				// 3.8.5.3 (2006/08/07)
	private String  refFilename		= null;				// 3.5.4.3 (2004/01/05)
	private String  refSheetName	= null;				// 3.5.4.3 (2004/01/05)
	private CreationHelper createHelper	= null;			// poi.xssf対応
	// 5.1.4.0 (2010/03/01) 行番号情報を、出力する(true)/しない(false)を指定
	private boolean		useNumber	= true;
	private int		skipRowCount	= 0;				// 6.0.1.2 (2014/08/08) データの書き込み開始位置設定

	/**
	 * DBTableModel から 各形式のデータを作成して,PrintWriter に書き出します。
	 * このメソッドは、EXCEL 書き出し時に使用します。
	 *
	 * 雛形シートを使う場合は、雛形シートの後ろに追加していきます。
	 * 最後には、雛形シートとして存在しな部分は、すべて削除します。
	 * 雛形シートを使わない場合は、新規シートに追加していくことになります。
	 * appendモードの場合は、オリジナルが雛形ファイルになります。
	 * 雛形ファイルの指定は、無効になります。
	 *
	 * @og.rev 4.0.0.0 (2006/09/31) 新規追加
	 * @og.rev 5.1.4.0 (2010/03/01) columns 対応 、useNumber属性対応
	 * @og.rev 6.0.1.2 (2014/08/08) ロジックの完全書き換え(保証もできない位･･･)
	 *
	 * @see #isExcel()
	 */
	@Override
	public void writeDBTable() {
//		if( ! createDBColumn() ) { return ; }		// 親クラスで実行

		useNumber = isUseNumber();

		// 3.5.6.0 (2004/06/18) 移動
		if( filename == null ) {
			String errMsg = "ファイルが指定されていません。";
			throw new HybsSystemException(errMsg );
		}

		// appendモードの場合は、オリジナルが雛形ファイルになります。
		// 雛形ファイルの指定は、無効になります。
		if( isAppend() ) {
			// 互換性のため、雛形ファイルが出力ファイルと同じでない場合はエラーとします。
			if( refFilename != null && !filename.equalsIgnoreCase( refFilename ) ) {
				String errMsg = "追加(appendモード)の時、雛形ファイルの指定は無効です。" + HybsSystem.CR
								+ " filename=[" + filename + "] , refFilename=[" + refFilename + "]"  + HybsSystem.CR ;
				throw new HybsSystemException( errMsg );
			}

			refFilename = filename ;
		}

		// 雛形があれば、雛形ファイルを読み込みます。なければ、オリジナルのファイルを読み込みます。
		String  nameUse = filename;
		boolean isRefFile = refFilename != null && refFilename.length() > 0;
		if( isRefFile ) {
			nameUse = refFilename ;
		}

		// 既存の場合は、WorkbookFactoryで作成。新規の場合は、拡張子(.xlsx/.xls)に応じて、作成する。
		Workbook wb = createWorkbook( nameUse , isRefFile );

		// 雛形ファイルのシート数。最後にこの分は削除します。
		int refCnt = isRefFile ? wb.getNumberOfSheets() : 0 ;

		// appendモードの場合は、オリジナルが雛形ファイルになるため、雛形シートの削除は行いません。
		if( isAppend() ) { refCnt = 0; }

		// 雛形シートを使う場合
		if( isRefFile ) {
			int refSheetIdx = 0;		// 雛形シート名の指定がない場合は、最初のシート
			if( refSheetName != null ) { refSheetIdx = wb.getSheetIndex(refSheetName); }

			if( refSheetIdx < 0 ) {		// 雛形シート名が存在しなかった。
				String errMsg = "指定の雛形シート名は、雛型ファイルに存在しませんでした。" + HybsSystem.CR
								+ " refFilename=[" + refFilename + "] , refSheetName=[" + refSheetName + "]"  + HybsSystem.CR ;
				throw new HybsSystemException( errMsg );
			}

			sheet = wb.cloneSheet( refSheetIdx );
		}
		else {
			// 3.8.6.0 (2006/08/07) フォント名やフォントサイズの指定
			// 雛形ファイルを使用しない場合のみ、設定可能です。
			Font font = wb.getFontAt((short)0);
			if( fontName != null ) {
				font.setFontName( fontName );	// "ＭＳ Ｐゴシック" など
			}
			if( fontPoint > 0 ) {
				font.setFontHeightInPoints( fontPoint );
			}

			sheet = wb.createSheet();
		}

		// シート名のセット。重複チェックと新しいシート名の作成処理を行う。
		wb.setSheetName(wb.getNumberOfSheets() -1, getNewSheetNameByName(wb, sheetName) );

		// poi.xssf対応(2009/04/08)
		createHelper = wb.getCreationHelper();

		nRowIndex = skipRowCount;		// 6.0.1.2 (2014/08/08) 開始行の指定

		super.writeDBTable( null );

		// 余計な行を削除
		removeSheetRow( sheet, nRowIndex );

		// 雛型ファイルを使っていた場合、その雛形シートを削除する
		if( isRefFile ) {
			for( int cnt=0; cnt<refCnt; cnt++ ) {
				wb.removeSheetAt( 0 );
			}
		}

		// メモリ中のデータをファイルに書き込む
		// 3.5.6.0 (2004/06/18) close を finally で処理するように変更。
		FileOutputStream fileOut = null ;
		try {
			fileOut = new FileOutputStream(filename);
			wb.write(fileOut);
		}
		catch( IOException ex ) {
			String errMsg = "ファイルへ書込み中にエラーが発生しました。" + HybsSystem.CR
						+ ex.getMessage() + HybsSystem.CR
						+ "  File=" + filename;				// 5.1.8.0 (2010/07/01) errMsg 修正
			throw new HybsSystemException( errMsg,ex );		// 3.5.5.4 (2004/04/15) 引数の並び順変更
		}
		finally {
			Closer.ioClose( fileOut );		// 4.0.0 (2006/01/31) close 処理時の IOException を無視
			if( null != sheet ) { sheet = null; }
			if( null != wb    ) { wb    = null; }
		}
	}

	/**
	 * DBTableModel から データを作成して,PrintWriter に書き出します。
	 *
	 * @og.rev 3.1.1.0 (2003/03/28) 同期メソッド(synchronized付き)を非同期に変更する。
	 * @og.rev 3.5.4.3 (2004/01/05) 引数に PrintWriter を受け取るように変更します。
	 * @og.rev 3.8.5.3 (2006/08/07) フォント名やフォントサイズの指定
	 * @og.rev 4.0.0.0 (2006/09/31) UnsupportedOperationException を発行します。
	 * @og.rev 6.0.1.2 (2014/08/08) 親クラスで同様の処理は定義済みのため、削除。
	 *
	 * @param	writer PrintWriterオブジェクト
	 */
//	@Override
//	public void writeDBTable( final PrintWriter writer )  {
//		String errMsg = "このクラスでは実装されていません。";
//		throw new UnsupportedOperationException( errMsg );
//	}

	/**
	 * PrintWriter に DBTableModelのラベル情報を書き込みます。
	 * 第一カラム目は、ラベル情報を示す "#Label" を書き込みます。
	 * この行は、出力形式に無関係に、TableWriter.TAB_SEPARATOR で区切られます。
	 *
	 * @og.rev 5.1.4.0 (2010/03/01) useNumber属性対応
	 * @og.rev 6.0.1.2 (2014/08/08) カラム飛ばしできる機能を追加
	 *
	 * @param	table  DBTableModelオブジェクト
	 * @param	writer PrintWriterオブジェクト
	 */
	@Override
	protected void writeLabel( final DBTableModel table,final PrintWriter writer ) {
		short nColIndex = 0;
		Row   oRow = setFirstCellValue( nRowIndex++, nColIndex++, "#Label" );
		for( int i=0; i<numberOfColumns; i++ ) {
			int clm = clmNo[i];
			if( clm < 0 ) {			// 6.0.1.2 (2014/08/08) カラム飛ばし
				setRowCellValue( oRow, nColIndex++, "", Cell.CELL_TYPE_STRING );
				continue;
			}

			String val = dbColumn[clm].getLabel();
			// 5.1.4.0 (2010/03/01) useNumber=false の場合は、nColIndex を一つ戻して、値に # を付ける。
			if( i==0 && !useNumber ) {
				nColIndex-- ;
				val = "#" + val;
			}
			setRowCellValue( oRow, nColIndex++, val,Cell.CELL_TYPE_STRING );
		}

		// 余計なセルを削除
		removeRowCell( oRow, nColIndex );
	}

	/**
	 * PrintWriter に DBTableModelの項目名情報を書き込みます。
	 * 第一カラム目は、項目名情報を示す "#Name" を書き込みます。
	 * この行は、出力形式に無関係に、TableWriter.TAB_SEPARATOR で区切られます。
	 *
	 * @og.rev 3.1.1.0 (2003/03/28) 同期メソッド(synchronized付き)を非同期に変更する。
	 * @og.rev 5.1.4.0 (2010/03/01) useNumber属性対応
	 * @og.rev 6.0.1.2 (2014/08/08) カラム飛ばしできる機能を追加
	 *
	 * @param	table  DBTableModelオブジェクト
	 * @param	writer PrintWriterオブジェクト
	 */
	@Override
	protected void writeName( final DBTableModel table,final PrintWriter writer ) {
		short nColIndex = 0;
		Row   oRow = setFirstCellValue( nRowIndex++, nColIndex++, "#Name" );
		for( int i=0; i<numberOfColumns; i++ ) {
			int clm = clmNo[i];
			if( clm < 0 ) {			// 6.0.1.2 (2014/08/08) カラム飛ばし
				setRowCellValue( oRow, nColIndex++, "",Cell.CELL_TYPE_STRING );
				continue;
			}

			String val = table.getColumnName(clm);
			// 5.1.4.0 (2010/03/01) useNumber=false の場合は、nColIndex を一つ戻して、値に # を付ける。
			if( i==0 && !useNumber ) {
				nColIndex-- ;
				val = "#" + val;
			}
			setRowCellValue( oRow, nColIndex++, val,Cell.CELL_TYPE_STRING );
		}

		// 余計なセルを削除
		removeRowCell( oRow, nColIndex );
	}

	/**
	 * PrintWriter に DBTableModelのサイズ情報を書き込みます。
	 * 第一カラム目は、サイズ情報を示す "#Size" を書き込みます。
	 * この行は、出力形式に無関係に、TableWriter.TAB_SEPARATOR で区切られます。
	 *
	 * @og.rev 3.5.5.5 (2004/04/23) DBColumn の size と maxlength の 意味を変更
	 * @og.rev 5.1.4.0 (2010/03/01) useNumber属性対応
	 * @og.rev 6.0.1.2 (2014/08/08) カラム飛ばしできる機能を追加
	 *
	 * @param	table  DBTableModelオブジェクト
	 * @param	writer PrintWriterオブジェクト
	 */
	@Override
	protected void writeSize( final DBTableModel table,final PrintWriter writer ) {
		short nColIndex = 0;
		Row   oRow = setFirstCellValue( nRowIndex++, nColIndex++, "#Size" );
		for( int i=0; i<numberOfColumns; i++ ) {
			int clm = clmNo[i];
			if( clm < 0 ) {			// 6.0.1.2 (2014/08/08) カラム飛ばし
				setRowCellValue( oRow, nColIndex++, "",Cell.CELL_TYPE_STRING );
				continue;
			}

			// 4.0.0 (2005/01/31) メソッド名変更
			String val = String.valueOf(dbColumn[clm].getTotalSize());
			// 5.1.4.0 (2010/03/01) useNumber=false の場合は、nColIndex を一つ戻して、値に # を付ける。
			if( i==0 && !useNumber ) {
				nColIndex-- ;
				val = "#" + val;
			}
			setRowCellValue( oRow, nColIndex++, val, Cell.CELL_TYPE_NUMERIC );
		}

		// 余計なセルを削除
		removeRowCell( oRow, nColIndex );
	}

	/**
	 * PrintWriter に DBTableModelのクラス名情報を書き込みます。
	 * 第一カラム目は、サイズ情報を示す "#Class" を書き込みます。
	 * この行は、出力形式に無関係に、TableWriter.TAB_SEPARATOR で区切られます。
	 *
	 * @og.rev 5.1.4.0 (2010/03/01) useNumber属性対応
	 * @og.rev 6.0.1.2 (2014/08/08) カラム飛ばしできる機能を追加
	 *
	 * @param	table  DBTableModelオブジェクト
	 * @param	writer PrintWriterオブジェクト
	 */
	@Override
	protected void writeClass( final DBTableModel table,final PrintWriter writer ) {
		short nColIndex = 0;
		Row   oRow = setFirstCellValue( nRowIndex++, nColIndex++, "#Class" );
		for( int i=0; i<numberOfColumns; i++ ) {
			int clm = clmNo[i];
			if( clm < 0 ) {			// 6.0.1.2 (2014/08/08) カラム飛ばし
				setRowCellValue( oRow, nColIndex++, "",Cell.CELL_TYPE_STRING );
				continue;
			}

			String val = dbColumn[clm].getClassName();
			// 5.1.4.0 (2010/03/01) useNumber=false の場合は、nColIndex を一つ戻して、値に # を付ける。
			if( i==0 && !useNumber ) {
				nColIndex-- ;
				val = "#" + val;
			}
			setRowCellValue( oRow, nColIndex++, val,Cell.CELL_TYPE_STRING );
		}

		// 余計なセルを削除
		removeRowCell( oRow, nColIndex );
	}

	/**
	 * PrintWriter に セパレーターを書き込みます。
	 * 第一カラム目は、サイズ情報を示す "#----" を書き込みます。
	 * この行は、出力形式に無関係に、TableWriter.TAB_SEPARATOR で区切られます。
	 *
	 * @og.rev 5.1.4.0 (2010/03/01) useNumber属性対応
	 *
	 * @param	table  DBTableModelオブジェクト
	 * @param	writer PrintWriterオブジェクト
	 */
	@Override
	protected void writeSeparator( final DBTableModel table,final PrintWriter writer ) {
		String sep = "----" ;

		short nColIndex = 0;
		Row   oRow = setFirstCellValue( nRowIndex++, nColIndex++, "#----" );
		for( int i=0; i<numberOfColumns; i++ ) {
			// 5.1.4.0 (2010/03/01) useNumber=false の場合は、単になにもしないだけ
			if( i==0 && !useNumber ) {
				continue;
			}
			setRowCellValue( oRow, nColIndex++, sep,Cell.CELL_TYPE_STRING );
		}

		// 余計なセルを削除
		removeRowCell( oRow, nColIndex );
	}

	/**
	 * PrintWriter に DBTableModelのテーブル情報を書き込みます。
	 * このクラスでは，データを ダブルコーテーション(")で囲みます。
	 * PrintWriter に DBTableModelのテーブル情報を書き込みます。
	 *
	 * @og.rev 3.8.0.1 (2005/06/17) DBTypeが NVAR の場合は、元のUnicodeに戻します。
	 * @og.rev 3.8.5.3 (2006/08/07) DBType の nativeType に対応した、CELL_TYPE をセットします。
	 * @og.rev 4.1.1.2 (2008/02/28) NativeタイプをEnum型(fukurou.model.NativeType)に変更
	 * @og.rev 5.1.4.0 (2010/03/01) columns 対応
	 * @og.rev 5.1.4.0 (2010/03/01) useNumber属性対応
	 * @og.rev 5.2.1.0 (2010/10/01) useRenderer 対応
	 * @og.rev 5.7.6.3 (2014/05/23) stringOutput対応
	 * @og.rev 6.0.1.2 (2014/08/08) カラム飛ばしできる機能を追加
	 *
	 * @param	table  DBTableModelオブジェクト
	 * @param	writer PrintWriterオブジェクト
	 */
	@Override
	protected void writeData( final DBTableModel table,final PrintWriter writer ) {
		int numberOfRows =	table.getRowCount();

		// 5.1.4.0 columns 対応。forループは、引数i で廻す。
		boolean[] nvar = new boolean[numberOfColumns];
		int[] cellType = new int[numberOfColumns];
		boolean useRenderer = isUseRenderer();	// 5.2.1.0 (2010/10/01)

		for( int i=0; i<numberOfColumns; i++ ) {
			int clm = clmNo[i];
			if( clm < 0 ) { continue; }			// 6.0.1.2 (2014/08/08) カラム飛ばし

			// 5.7.6.3 (2014/05/23) ここでレンデラ時の文字タイプ判定を行う
			if( useRenderer && dbColumn[clm].isStringOutput() ){
				cellType[i] = Cell.CELL_TYPE_STRING ;
			}
			else {
				NativeType nativeType = dbColumn[clm].getNativeType();
				switch( nativeType ) {
					case INT    :
					case LONG   :
					case DOUBLE :
						cellType[i] = Cell.CELL_TYPE_NUMERIC ;
							break;
					case STRING :
					case CALENDAR :
					default :
							cellType[i] = Cell.CELL_TYPE_STRING ;
							break;
				}
			}
			nvar[i] = "NVAR".equals( dbColumn[clm].getDbType()) ;
		}
//		boolean useRenderer = isUseRenderer();	// 5.2.1.0 (2010/10/01) , 5.7.6.3 (2014/05/23) 上へ移動

		short nColIndex;
		Row   oRow;
		for( int row=0; row<numberOfRows; row++ ) {
			nColIndex = 0;
			oRow = setFirstCellValue( nRowIndex++, nColIndex++, String.valueOf( row+1 ) );

			// 5.1.4.0 (2010/03/01) useNumber=false の場合は、nColIndex を一つ戻す。
			if( !useNumber ) {
				nColIndex-- ;
			}

			for( int i=0; i<numberOfColumns; i++ ) {
				int clm = clmNo[i];
				if( clm < 0 ) {			// 6.0.1.2 (2014/08/08) カラム飛ばし
					setRowCellValue( oRow, nColIndex++, "",Cell.CELL_TYPE_STRING );
					continue;
				}

				String val = table.getValue(row,clm);
				if( nvar[i] ) {
					val = StringUtil.getReplaceEscape( val );
				}
				// 5.2.1.0 (2010/10/01) useRenderer 対応
				else if( useRenderer ) {
					val = StringUtil.spanCut( dbColumn[clm].getRendererValue( val ) );
				}

				setRowCellValue( oRow, nColIndex++, val,cellType[i] );
			}

			// 余計なセルを削除
			removeRowCell( oRow, nColIndex );
		}
	}

	/**
	 * Excelの指定行の一番目セルにデータを設定する。
	 *
	 * @og.rev 4.3.4.0 (2008/12/01) POI3.2対応
	 * @og.rev 4.3.4.3 (2008/12/22) protected化
	 * @og.rev 6.0.1.2 (2014/08/08) private化
	 *
	 * @param	 indexOfRow 行の番号
	 * @param	 indexOfCell セルの番号
	 * @param	 dataVal    String文字列
	 *
	 * @return	 Rowのobject型
	 */
//	protected Row setFirstCellValue( final int indexOfRow, final int indexOfCell, final String dataVal ) {
	private Row setFirstCellValue( final int indexOfRow, final int indexOfCell, final String dataVal ) {
		Row oRow = sheet.getRow( indexOfRow );
		if( oRow == null ) { oRow = sheet.createRow( indexOfRow ); }

		Cell oCell = oRow.getCell( indexOfCell );
		if( null == oCell ) { oCell = oRow.createCell( indexOfCell ); }

		RichTextString richText = createHelper.createRichTextString( dataVal );
		oCell.setCellValue( richText );

		return oRow;
	}

	/**
	 * Excelの指定セルにデータを設定する。
	 *
	 * @og.rev 4.3.4.0 (2008/12/01) POI3.2対応
	 * @og.rev 4.3.4.3 (2008/12/22) protected化
	 * @og.rev 5.7.4.1 (2014/03/15) useRenderer対応
	 * @og.rev 5.7.6.3 (2014/05/23) stringOutput対応
	 * @og.rev 6.0.1.2 (2014/08/08) private化
	 *
	 * @param	 oRow   Row型のオブジェクト
	 * @param	 indexOfCell セルの番号
	 * @param	 dataVal    String文字列
	 * @param	 cellType   [Cell.CELL_TYPE_STRING/Cell.CELL_TYPE_NUMERIC]
	 */
//	protected void setRowCellValue( final Row oRow, final int indexOfCell, final String dataVal,final int cellType ) {
	private void setRowCellValue( final Row oRow, final int indexOfCell, final String dataVal,final int cellType ) {
		Cell oCell = oRow.getCell( indexOfCell );
		if( null == oCell ) { oCell = oRow.createCell( indexOfCell ); }

		// 5.7.4.1 (2014/03/15) useRendererがtrueの場合はdouble変換をかけない
//		if( !isUseRenderer() && cellType == Cell.CELL_TYPE_NUMERIC ) {
		if( cellType == Cell.CELL_TYPE_NUMERIC ) {
			oCell.setCellValue( StringUtil.parseDouble( dataVal ) );
		}
		else {
			RichTextString richText = createHelper.createRichTextString( dataVal );
			oCell.setCellValue( richText );
		}
	}

	/**
	 * Excelの指定セルをシートから削除する。
	 *
	 * @og.rev 4.3.4.0 (2008/12/01) POI3.2対応
	 * @og.rev 4.3.4.3 (2008/12/22) protected化
	 * @og.rev 6.0.1.2 (2014/08/08) private化
	 *
	 * @param	 oRow   Row型のオブジェクト
	 * @param	 nBegin     セルの開始番号
	 */
//	protected void removeRowCell( final Row oRow, final int nBegin ) {
	private void removeRowCell( final Row oRow, final int nBegin ) {
		int nEnd = oRow.getLastCellNum();
//		for( int nIndex = nBegin; nIndex <= nEnd; nIndex++) {
		for( int nIndex = nEnd; nIndex >= nBegin; nIndex--) {
			Cell oCell = oRow.getCell( nIndex );
			if( null != oCell ) { oRow.removeCell(oCell); }
		}
	}

	/**
	 * Excelの指定行をシートから削除する。
	 * nBegin は、すでに＋１されているので、＝ になるまで削除します。
	 *
	 * @param	 oThisSheet Sheet型のオブジェクト
	 * @param	 nBegin     行の開始番号
	 */
	private void removeSheetRow( final Sheet oThisSheet, final int nBegin ) {
		int nEnd = oThisSheet.getLastRowNum();
//		for( int nIndex = nBegin; nIndex <= nEnd; nIndex++) {
		for( int nIndex = nEnd; nIndex >= nBegin; nIndex--) {
			Row oRow = oThisSheet.getRow( nIndex );
			if( null != oRow ) { oThisSheet.removeRow(oRow); }
		}
	}

	/**
	 * DBTableModelのデータとして書き込むときのシート名をセットします。
	 * 初期値は、"Sheet1" です。同じ名称のシートが存在する場合、そのシート
	 * 名の後に(2)、(3)のような文字列をくっ付けます。
	 *
	 * ※ 改造前は、シートの名前は30桁文字(31個文字)なので、そのチェックと
	 *    substring による桁数調整が入っていましたが、やめます。
	 *
	 * @og.rev 6.0.1.2 (2014/08/08) シート名の桁数チェックを廃止(簡略化と希望的観測)
	 *
	 * @param	 workbook   Workbookオブジェクト
	 * @param	 nameSet    String文字列,指定のシート名
	 * @return	シート名
	 */
//	protected String getNewSheetNameByName( final Workbook workbook, final String nameSet) {
	private String getNewSheetNameByName( final Workbook workbook, final String nameSet) {
		String nameSheet = nameSet;
		int nCount  = 1;

		while( workbook.getSheetIndex(nameSheet) >= 0 ) {	// シート名が存在している場合
			nameSheet = nameSet + "(" + nCount + ")";
			nCount++;
		}

		return nameSheet;
	}

	/**
	 * 出力先ファイル名をセットします。(DIR + Filename)
	 * これは、EXCEL追加機能として実装されています。
	 *
	 * @og.rev 3.5.4.3 (2004/01/05) 新規作成
	 *
	 * @param   fname EXCEL雛型参考ファイル名
	 */
	@Override
	public void setFilename( final String fname ) {
		filename = fname;
	}

	/**
	 * DBTableModelのデータとして読み込むときのシート名を設定します。
	 * 初期値は、"Sheet1" です。
	 * これは、EXCEL追加機能として実装されています。
	 *
	 * @og.rev 3.5.4.2 (2003/12/15) 新規追加
	 *
	 * @param   shtName シート名
	 */
	@Override
	public void setSheetName( final String shtName ) {
		if( shtName != null ) { this.sheetName = shtName; }
	}

	/**
	 * このクラスが、EXCEL対応機能を持っているかどうかを返します。
	 *
	 * EXCEL対応機能とは、シート名のセット、雛型参照ファイル名のセット、
	 * 書き込み元ファイルのFileオブジェクト取得などの、特殊機能です。
	 * 本来は、インターフェースを分けるべきと考えますが、taglib クラス等の
	 * 関係があり、問い合わせによる条件分岐で対応します。
	 *
	 * @og.rev 3.5.4.3 (2004/01/05) 新規追加
	 *
	 * @return	EXCEL対応機能を持っているかどうか(常に true)
	 */
	@Override
	public boolean isExcel() {
		return true;
	}

	/**
	 * EXCEL雛型参考ファイル名をセットします。(DIR + Filename)
	 * これは、EXCEL追加機能として実装されています。
	 *
	 * @og.rev 3.5.4.3 (2004/01/05) 新規作成
	 *
	 * @param   rfname EXCEL雛型参考ファイル名
	 */
	@Override
	public void setRefFilename( final String rfname ) {
		refFilename = rfname;
	}

	/**
	 * EXCEL雛型参考ファイルのシート名を設定します。
	 * これは、EXCEL追加機能として実装されています。
	 *
	 * EXCELファイルを書き出す時に、雛型として参照するシート名を指定します。
	 * これにより、複数の形式の異なるデータを順次書き出したり(appendモードを併用)する
	 * ことや、シートを指定して新規にEXCELを作成する場合にフォームを設定する事が可能になります。
	 * 初期値は、null(第一シート) です。
	 *
	 * @og.rev 3.5.4.3 (2004/01/05) 新規追加
	 *
	 * @param   rshtName シート名
	 */
	@Override
	public void setRefSheetName( final String rshtName )  {
		if( rshtName != null ) { refSheetName = rshtName; }
	}

	/**
	 * データの書き込み開始位置を設定します(初期値:0)。
	 *
	 * TAB区切りテキストやEXCEL等のデータの書き込みの開始位置を指定します。
	 * 属性名は、行を飛ばす処理ということで、readTable タグと同じ名称です。
	 * ファイルの先頭行が、０行としてカウントしますので、設定値は、読み飛ばす
	 * 件数になります。(１と指定すると、１件読み飛ばし、２行目から読み込みます。)
	 * 行の読み飛ばしと、カラムの読み飛ばし(columns)、refFileURL、refFilename、
	 * refSheetName とともに使用すれば、ある程度のレイアウト設定が可能です。
	 * なお、この機能は、TableWriter_Excel のみに実装します。
	 *
	 * @og.rev 6.0.1.2 (2014/08/08) 新規作成
	 *
	 * @param	skipCount 書き込み開始位置
	 */
	@Override
	public void setSkipRowCount( final int skipCount ) {
		skipRowCount = skipCount ;
	}

	/**
	 * EXCEL出力時のデフォルトフォント名を設定します。
	 * これは、EXCEL追加機能として実装されています。
	 *
	 * EXCELファイルを書き出す時に、デフォルトフォント名を指定します。
	 * フォント名は、EXCELのフォント名をそのまま使用してください。
	 * 内部的に、POI の org.apache.poi.hssf.usermodel.HSSFFont#setFontName( String )
	 * に設定されます。
	 * 雛形ファイルを使用しない場合のみ、設定可能です。
	 *
	 * @og.rev 3.8.5.3 (2006/08/07) 新規追加
	 *
	 * @param   font デフォルトフォント名
	 */
	@Override
	public void setFontName( final String font ) {
		fontName = font ;
	}

	/**
	 * EXCEL出力時のデフォルトフォントポイント数を設定します。
	 * これは、EXCEL追加機能として実装されています。
	 *
	 * EXCELファイルを書き出す時に、デフォルトポイント数を指定します。
	 * 内部的に、POI の org.apache.poi.hssf.usermodel.HSSFFont#setFontHeightInPoints( short )
	 * に設定されます。
	 * 雛形ファイルを使用しない場合のみ、設定可能です。
	 *
	 * @og.rev 3.8.5.3 (2006/08/07) 新規追加
	 *
	 * @param   point デフォルトフォントポイント数
	 */
	@Override
	public void setFontPoint( final short point ) {
		fontPoint = point;
	}

	/**
	 * EXCELファイルのWookbookというStream(MicrosoftのOLE用語)を作ります
	 * 条件によって、新規かとファイルから読み込み書き込みかが分かれます。
	 *
	 * @og.rev 6.0.1.2 (2014/08/08) メソッド統合
	 *
	 * @param   fname  EXCEL参考ファイル名
	 * @param   isOpen true:ファイルオープン/false:新規作成
	 *
	 * @return  EXCELファイルのWorkbook
	 */
//	protected Workbook createWorkbook( final String fname ) {
	private Workbook createWorkbook( final String fname , final boolean isOpen ) {
		final Workbook rtnWb ;

		if( isOpen ) {
			// File オブジェクトでcreate すると、ファイルがオープンされたままになってしまう。
			FileInputStream fileIn  = null;
			try {
				fileIn = new FileInputStream(fname);
				rtnWb = WorkbookFactory.create(fileIn);
			}
			catch ( Exception ex ) {
				String errMsg = "ファイル読込みエラー[" + fname + "]"  ;
				throw new HybsSystemException( errMsg,ex );		// 3.5.5.4 (2004/04/15) 引数の並び順変更
			}
			finally {
				Closer.ioClose( fileIn );		// 4.0.0 (2006/01/31) close 処理時の IOException を無視
			}
		}
		else {
			// 新規の場合、ファイル名に.xlsxで終了した場合⇒.xlsx形式ファイル作成、その他⇒.xls形式ファイル作成
			if( fname.toLowerCase(Locale.JAPAN).endsWith( ".xlsx" ) ) {
				rtnWb = new org.apache.poi.xssf.usermodel.XSSFWorkbook();
			}
			else {
				rtnWb = new org.apache.poi.hssf.usermodel.HSSFWorkbook();
			}
		}
		return rtnWb;
	}
}
